"""
.. topic:: ``aws.py``

    AWS helper functions.
"""

import boto3
import requests
from botocore.exceptions import ClientError

from infrahouse_toolkit import LOG

VALUE_MAP = {
    "AWS_ACCESS_KEY_ID": "AccessKeyId",
    "AWS_SECRET_ACCESS_KEY": "SecretAccessKey",
    "AWS_SESSION_TOKEN": "SessionToken",
    # These are old s3.fs options
    # Soon they will be deprecated
    # https://github.com/s3fs-fuse/s3fs-fuse/pull/1729
    "AWSACCESSKEYID": "AccessKeyId",
    "AWSSECRETACCESSKEY": "SecretAccessKey",
    "AWSSESSIONTOKEN": "SessionToken",
}


def assume_role(role_arn, region=None) -> dict:
    """
    Assume a given role and return a dictionary with credentials.

    :param role_arn: Role to be assumed.
    :type role_arn: str
    :param region: AWS region name.
    :type region: str
    :return: A dictionary with three keys: AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, and
    """
    try:
        LOG.debug("Assuming role %s", role_arn)
        client = boto3.client("sts", region_name=region)
        response = client.assume_role(RoleArn=role_arn, RoleSessionName="ih-s3-reprepro-s3fs")
        LOG.debug("Got credentials %r", response["Credentials"])
        return {var: response["Credentials"].get(key) for var, key in VALUE_MAP.items()}
    except ClientError as err:
        LOG.error(err)
        LOG.debug("To revert environment:\n%s", "\n".join([f"unset {key}" for key in VALUE_MAP]))
        raise


def get_client(service_name, role_arn=None, region=None, session_name=__name__):
    """
    Get an AWS service client assuming a role if specified.

    :param service_name: AWS service. ec2, sts, etc.
    :type service_name: str
    :param role_arn: Role ARN if it needs to be assumed.
    :type role_arn: str
    :param session_name: A human-readable string that tells something about this session.
        Exact value isn't important.
    :type session_name: str
    :param region: AWS region name.
    :type region: str
    :return: AWS boto3 client.
    """
    if role_arn:
        sts = boto3.client("sts", region_name=region)
        aws_iam_role = sts.assume_role(RoleArn=role_arn, RoleSessionName=session_name)
        session = boto3.Session(
            aws_access_key_id=aws_iam_role["Credentials"]["AccessKeyId"],
            aws_secret_access_key=aws_iam_role["Credentials"]["SecretAccessKey"],
            aws_session_token=aws_iam_role["Credentials"]["SessionToken"],
            region_name=region,
        )
        sts = session.client("sts", region_name=region)
        response = sts.get_caller_identity()
        LOG.debug("Assumed role: %s", response)
        return session.client(service_name, region_name=region)

    return boto3.client(service_name, region_name=region)


def get_credentials_from_profile() -> dict:
    """
    Another way to get AWS credentials is from EC2 instance metadata.

    :return: A dictionary with AWS_* variables.
    """
    url = "http://169.254.169.254/latest/meta-data/iam/security-credentials/"
    profile_name = requests.get(url, timeout=10).text
    profile_data = requests.get(f"{url}/{profile_name}", timeout=10).json()
    profile_data["SessionToken"] = profile_data["Token"]

    return {var: profile_data.get(key) for var, key in VALUE_MAP.items()}



"""
:py:mod:`infrahouse_toolkit.terraform.backend` exceptions.
"""

from infrahouse_toolkit.terraform.exceptions import IHTFException


class IHBackendException(IHTFException):
    """Terraform backend exceptions."""


class IHUnknownBackend(IHBackendException):
    """Not supported Terraform backend."""



from os import path as osp

from infrahouse_toolkit.terraform import parse_plan
from infrahouse_toolkit.terraform.backends.s3backend import TFS3Backend
from infrahouse_toolkit.terraform.status import RunOutput, RunResult, TFStatus


def test_summary_resources():
    counts, resources = parse_plan(
        open(osp.join(osp.dirname(osp.realpath(__file__)), "../plans/plan-0-0-0.stdout")).read(),
    )
    status = TFStatus(
        TFS3Backend("foo_backet", "path/to/tf.state"),
        True,
        counts,
        RunOutput("no stdout", "no stderr"),
        affected_resources=resources,
    )
    print("")
    print(status.summary_resources)



"""
.. topic:: ``utils.py``

    Various helper functions.
"""
import json
import sys
from contextlib import contextmanager
from os import getgid, getuid
from os import path as osp
from subprocess import CalledProcessError, Popen, check_call
from tempfile import TemporaryDirectory
from time import sleep, time

import boto3

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING, LOG
from infrahouse_toolkit.cli.ih_s3_reprepro.aws import (
    assume_role,
    get_credentials_from_profile,
)
from infrahouse_toolkit.cli.ih_s3_reprepro.gpg import gpg

DEPENDENCIES = ["reprepro", "gpg", "s3fs"]


def check_dependencies(binaries: list):
    """
    Ensure that dependencies are installed. The function calls each of the binary
    with a ``--help`` option.

    :param binaries: Dependency commands. List of strings.
    :type binaries: list
    """
    for dep in binaries:
        try:
            with open("/dev/null", "w", encoding=DEFAULT_OPEN_ENCODING) as devnull:
                LOG.debug("Checking if %s is installed.", dep)
                check_call([dep, "--help"], stdout=devnull, stderr=devnull)
        except FileNotFoundError:
            LOG.error("Looks like %s is not installed", dep)
            LOG.info("Try installing it by \n\n\tapt-get install %s\n", dep)
            sys.exit(1)


def mount_s3(bucket: str, path: str, role_arn: str = None, region: str = None):
    """
    Mount an S3 bucket at a path.

    :param bucket: AWS S3 bucket name.
    :type bucket: str
    :param path: Local filesystem path name.
    :type path: str
    :param role_arn: Assume role if specified.
    :param region: AWS region name.
    :type region: str
    """
    env = {}
    cmd = ["s3fs", bucket, path, "-o", f"uid={getuid()}", "-o", f"gid={getgid()}"]
    if role_arn:
        env = assume_role(role_arn)
        sts = boto3.client(
            "sts",
            aws_access_key_id=env["AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=env["AWS_SECRET_ACCESS_KEY"],
            aws_session_token=env["AWS_SESSION_TOKEN"],
            region_name=region,
        )
        response = sts.get_caller_identity()
        LOG.debug("Assumed role: %s", response)
    else:
        env = get_credentials_from_profile()

    LOG.debug("To reproduce environment: \n%s", "\n".join([f'export {key}="{value}"' for key, value in env.items()]))
    LOG.debug("Command to debug: mkdir -p %s; %s -o dbglevel=info -f -o curldbg", path, " ".join(cmd))
    execute(cmd, env=env)


def umount_s3(path: str):
    """
    Unmount an S3 bucket at a path.

    :param path: Local filesystem path name where the S3 bucket is mounted at.
    :type path: str
    """
    try:
        check_call(["umount", path])
    except CalledProcessError as err:
        LOG.error(err)
        sys.exit(1)


@contextmanager
def local_s3(bucket, role_arn=None, retry_timeout=60, region=None) -> str:
    """
    Mount an S3 bucket locally and return a mount point.

    :param bucket: AWS S3 bucket name.
    :type bucket: str
    :param role_arn: Assume role if specified.
    :type role_arn: str
    :param retry_timeout: How many second to keep trying to mount the bucket.
    :type retry_timeout: int
    :param region: AWS region name.
    :type region: str
    :return: Local filesystem path where the S3 bucket is mounted at.
    """
    with TemporaryDirectory() as mnt_dir:
        try:
            now = time()
            timeout = now + retry_timeout
            while True:
                mount_s3(bucket, mnt_dir, role_arn=role_arn, region=region)
                if osp.exists(osp.join(mnt_dir, "conf/distributions")):
                    break
                LOG.warning("Waiting until s3://%s is mounted at %s", bucket, mnt_dir)
                sleep(1)
                if time() > timeout:
                    raise RuntimeError(f"s3://{bucket} is not mounted after {retry_timeout} seconds")
            yield mnt_dir

        finally:
            umount_s3(mnt_dir)


def execute(cmd: list, env: dict = None):
    """
    Execute a command and exit with 1 if the command raises CalledProcessError.

    :param cmd: A command to execute. It's passed to check_call() and therefore must be a list.
    :type cmd: list
    :param env: Pass a dictionary with environment
    :type env: dict
    """
    try:
        LOG.debug("Executing %s", " ".join(cmd))
        LOG.debug("Environment: %s", json.dumps(env, indent=4) if env else "None")
        with Popen(cmd, env=env) as proc:
            try:
                proc.communicate()
            except KeyboardInterrupt:
                LOG.info("Exiting on <ctrl>+c")
                proc.terminate()
                LOG.info("Process %s is terminated. Waiting for it to exit.", cmd[0])
                proc.wait(60)

    except CalledProcessError as err:
        LOG.exception(err)
        sys.exit(1)


@contextmanager
def repo_env(bucket, role_arn, gpg_key_secret_id, gpg_passphrase_secret_id, region=None):
    """
    Prepare locally a repo and GPG so "reprepro" can manage it.

    :param bucket: AWS S3 bucket with the repo. The repo must be in the root.
    :type bucket: str
    :param role_arn: Optional role ARN. If specified, AWS client will assume it.
    :type role_arn: str
    :param gpg_key_secret_id: AWS secretsmanager secret (name or ARN) that stores the private GPG key,
        needed by ``reprepro`` to sign the repo.
    :type gpg_key_secret_id: str
    :param gpg_passphrase_secret_id: AWS secretsmanager secret (name or ARN) that stores a passphrase for
        the private GPG key. Note, it's not the passphrase itself, it's a secret that stores it.
    :type gpg_passphrase_secret_id: str
    :param region: AWS region name.
    :type region: str
    :return: A tuple with two strings: Local filesystem directory with a mounted S3 bucket
        and GPG home directory.
    """
    with local_s3(bucket, role_arn, region=region) as path:
        with gpg(
            secret_key=gpg_key_secret_id, role_arn=role_arn, secret_passphrase=gpg_passphrase_secret_id, region=region
        ) as gpg_home:
            yield path, gpg_home



"""
.. topic:: ``ih-plan``

    A group of commands to work with a Terraform plan file.
    The command can upload the plan to an S3 bucket, download it, and remove.

    See ``ih-plan --help`` for more details.
"""
import click

from infrahouse_toolkit.cli.ih_plan.cmd_download import cmd_download
from infrahouse_toolkit.cli.ih_plan.cmd_min_permissions import cmd_min_permissions
from infrahouse_toolkit.cli.ih_plan.cmd_publish import cmd_publish
from infrahouse_toolkit.cli.ih_plan.cmd_remove import cmd_remove
from infrahouse_toolkit.cli.ih_plan.cmd_upload import cmd_upload
from infrahouse_toolkit.cli.lib import DEFAULT_TF_BACKEND_FILE


@click.group()
@click.option(
    "--bucket",
    help="AWS S3 bucket name to upload/download the plan. "
    "By default, parse Terraform backend configuration (see --tf-backend-file)"
    " in the current directory.",
    default=None,
)
@click.option(
    "--aws-assume-role-arn",
    help="ARN of a role the AWS client should assume.",
    default=None,
)
@click.option(
    "--tf-backend-file",
    help="File with Terraform backend configuration.",
    default=DEFAULT_TF_BACKEND_FILE,
    show_default=True,
)
@click.version_option()
@click.pass_context
def ih_plan(ctx, bucket, aws_assume_role_arn, tf_backend_file):
    """Terraform plan helpers."""
    ctx.obj = {"aws_assume_role_arn": aws_assume_role_arn, "bucket": bucket, "tf_backend_file": tf_backend_file}


for cmd in [cmd_upload, cmd_download, cmd_remove, cmd_publish, cmd_min_permissions]:
    # noinspection PyTypeChecker
    ih_plan.add_command(cmd)



from infrahouse_toolkit.terraform.backends.s3backend import TFS3Backend
from infrahouse_toolkit.terraform.status import RunOutput, RunResult, TFStatus


def test_summary_counts():
    status = TFStatus(
        TFS3Backend("foo_backet", "path/to/tf.state"),
        True,
        RunResult(1, 1, 1),
        RunOutput("no stdout", "no stderr"),
        affected_resources=RunResult(["a"], ["b"], ["c"]),
    )
    print("")
    print(status.summary_counts)
    status.success = False
    print("")
    print(status.summary_counts)



from os import environ

import pytest

from infrahouse_toolkit.terraform import RunOutput, TFStatus, parse_comment, parse_plan
from infrahouse_toolkit.terraform.backends import TFS3Backend
from infrahouse_toolkit.terraform.githubpr import GitHubPR


@pytest.mark.skipif("GITHUB_TOKEN" not in environ, reason="This is a development test, needs GITHUB_TOKEN")
def test_publish_comment():
    gh_pr = GitHubPR("infrahouse8/github-control", 33)
    gh_pr.delete_my_comments()
    counts, resources = parse_plan(open("infrahouse_toolkit/terraform/tests/plans/plan-2-1-2.stdout").read())
    status = TFStatus(
        TFS3Backend("foo-bucket", "path/to/key.state"),
        success=True,
        run_result=counts,
        run_output=RunOutput(open("infrahouse_toolkit/terraform/tests/plans/plan-2-1-2.stdout").read(), None),
        affected_resources=resources,
    )
    print(status.comment)
    gh_pr.publish_comment(status.comment)



"""
.. topic:: ``ih-plan remove``

    A ``ih-plan remove`` subcommand.

    See ``ih-plan remove --help`` for more details.
"""

import click

from infrahouse_toolkit.cli.lib import get_bucket, get_s3_client


def validate_key_name(ctx, param, value):  # pylint: disable=unused-argument
    """Check if passed value ends with a ".state"."""
    bad_ending = ".state"
    if value.endswith(bad_ending):
        raise click.BadParameter(f"The file cannot end with a {bad_ending}")

    return value


@click.command(name="remove")
@click.argument("key_name", callback=validate_key_name)
@click.pass_context
def cmd_remove(ctx, key_name):
    """
    Remove a file from an S3 bucket.

    It could be a simple ``aws s3 rm s3://...`` but the command will do two things additionally:

    \b
    - It will parse a Terraform backend configuration to find the bucket name (See ih-plan --help).
    - It will refuse to delete a file if it ends with a ".state".
    """
    s3_client = get_s3_client(role=ctx.obj["aws_assume_role_arn"])
    bucket = ctx.obj["bucket"] or get_bucket(ctx.obj["tf_backend_file"])
    s3_client.delete_object(Bucket=bucket, Key=key_name)
    print(f"Successfully removed s3://{bucket}/{key_name}.")



"""
.. topic:: ``ih-plan min-permissions``

    A ``ih-plan min-permissions`` subcommand.

    See ``ih-plan min-permissions --help`` for more details.
"""
import json
from copy import deepcopy
from json import JSONDecodeError

import click

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING


@click.command(name="min-permissions")
@click.option("--existing-actions", help="A file with permissions.", default=None)
@click.argument("trace_file")
def cmd_min_permissions(existing_actions, trace_file):
    """
    Parse Terraform trace file and produce an action list from the trace.

    The trace file contains entries with AWS actions. The command
    finds the actions and AWS services to generate a list that
    you can add to an AWS policy.
    It's useful to prepare the least privileges policy.

    The output looks similar to this:

    \b
    [
        "ec2:DeleteNatGateway",
        "ec2:DescribeAddresses",
        "ec2:DescribeInternetGateways",
        "ec2:DescribeNatGateways",
    ]

    """
    actions = []
    if existing_actions:
        with open(existing_actions, encoding=DEFAULT_OPEN_ENCODING) as f_desc:
            actions = json.loads(f_desc.read())

    old_actions = deepcopy(actions)
    initial_count = len(old_actions)

    print(f"## Existing {initial_count} actions:")
    print(json.dumps(sorted(old_actions), indent=4))

    with open(trace_file, encoding=DEFAULT_OPEN_ENCODING) as f_decs:
        for line in f_decs.readlines():
            try:
                operation = json.loads(line)
                if "aws.operation" in operation:
                    actions.append(f'{operation["aws.service"].lower()}:{operation["aws.operation"]}')
            except JSONDecodeError:
                pass

    new_actions = list(set(actions))
    final_count = len(new_actions)
    print(f"## {final_count - initial_count} new action(s):")
    print(json.dumps([x for x in new_actions if x not in old_actions], indent=4))

    print(f"## Old and new actions together, {final_count} in total:")
    print(json.dumps(sorted(list(set(actions))), indent=4))
